@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1999, 2022, Oracle and/or its affiliates. All rights reserved.
+ * Copyright (c) 1999, 2023, Oracle and/or its affiliates. All rights reserved.
  * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
  *
  * This code is free software; you can redistribute it and/or modify it
@@ -105,8 +105,8 @@ public:
 
   // Returns previous value.
   template<typename D, typename I>
-  inline static D fetch_and_add(D volatile* dest, I add_value,
-                                atomic_memory_order order = memory_order_conservative);
+  inline static D fetch_then_add(D volatile* dest, I add_value,
+                                 atomic_memory_order order = memory_order_conservative);
 
   template<typename D, typename I>
   inline static D sub(D volatile* dest, I sub_value,
@@ -151,7 +151,7 @@ public:
                           T exchange_value,
                           atomic_memory_order order = memory_order_conservative);
 
-  // Performs atomic compare of *dest and NULL, and replaces *dest
+  // Performs atomic compare of *dest and nullptr, and replaces *dest
   // with exchange_value if the comparison succeeded.  Returns true if
   // the comparison succeeded and the exchange occurred.  This is
   // often used as part of lazy initialization, as a lock-free
@@ -237,7 +237,6 @@ public:
   }
 
 private:
-WINDOWS_ONLY(public:) // VS2017 warns (C2027) use of undefined type if IsPointerConvertible is declared private
   // Test whether From is implicitly convertible to To.
   // From and To must be pointer types.
   // Note: Provides the limited subset of C++11 std::is_convertible
@@ -320,14 +319,14 @@ private:
   // - platform_add is an object of type PlatformAdd<sizeof(D)>.
   //
   // Then both
-  //   platform_add.add_and_fetch(dest, add_value, order)
-  //   platform_add.fetch_and_add(dest, add_value, order)
+  //   platform_add.add_then_fetch(dest, add_value, order)
+  //   platform_add.fetch_then_add(dest, add_value, order)
   // must be valid expressions returning a result convertible to D.
   //
-  // add_and_fetch atomically adds add_value to the value of dest,
+  // add_then_fetch atomically adds add_value to the value of dest,
   // returning the new value.
   //
-  // fetch_and_add atomically adds add_value to the value of dest,
+  // fetch_then_add atomically adds add_value to the value of dest,
   // returning the old value.
   //
   // When the destination type D of the Atomic operation is a pointer type P*,
@@ -339,7 +338,7 @@ private:
   // 1, casting if needed.  It also scales add_value by sizeof(P).  The result
   // of the platform operation is cast back to P*.  This means the platform
   // operation does not need to account for the scaling.  It also makes it
-  // easy for the platform to implement one of add_and_fetch or fetch_and_add
+  // easy for the platform to implement one of add_then_fetch or fetch_then_add
   // in terms of the other (which is a common approach).
   //
   // No definition is provided; all platforms must explicitly define
@@ -399,11 +398,15 @@ private:
                                 T compare_value,
                                 T exchange_value);
 
-  // Support platforms that do not provide Read-Modify-Write
-  // byte-level atomic access. To use, derive PlatformCmpxchg<1> from
-  // this class.
+  // Support platforms that do not provide Read-Modify-Write atomic
+  // accesses for 1-byte and 8-byte widths. To use, derive PlatformCmpxchg<1>,
+  // PlatformAdd<S>, PlatformXchg<S> from these classes.
 public: // Temporary, can't be private: C++03 11.4/2. Fixed by C++11.
   struct CmpxchgByteUsingInt;
+  template<size_t byte_size>
+  struct XchgUsingCmpxchg;
+  template<size_t byte_size>
+  class AddUsingCmpxchg;
 private:
 
   // Dispatch handler for xchg.  Provides type-based validity
@@ -678,6 +681,47 @@ struct Atomic::CmpxchgByteUsingInt {
                atomic_memory_order order) const;
 };
 
+// Define the class before including platform file, which may use this
+// as a base class, requiring it be complete.  The definition is later
+// in this file, near the other definitions related to xchg.
+template<size_t byte_size>
+struct Atomic::XchgUsingCmpxchg {
+  template<typename T>
+  T operator()(T volatile* dest,
+               T exchange_value,
+               atomic_memory_order order) const;
+};
+
+// Define the class before including platform file, which may use this
+// as a base class, requiring it be complete.
+template<size_t byte_size>
+class Atomic::AddUsingCmpxchg {
+public:
+  template<typename D, typename I>
+  static inline D add_then_fetch(D volatile* dest,
+                                 I add_value,
+                                 atomic_memory_order order) {
+    D addend = add_value;
+    return fetch_then_add(dest, add_value, order) + add_value;
+  }
+
+  template<typename D, typename I>
+  static inline D fetch_then_add(D volatile* dest,
+                          I add_value,
+                          atomic_memory_order order) {
+    STATIC_ASSERT(byte_size == sizeof(I));
+    STATIC_ASSERT(byte_size == sizeof(D));
+
+    D old_value;
+    D new_value;
+    do {
+      old_value = Atomic::load(dest);
+      new_value = old_value + add_value;
+    } while (old_value != Atomic::cmpxchg(dest, old_value, new_value, order));
+    return old_value;
+  }
+};
+
 // Define the class before including platform file, which may specialize
 // the operator definition.  No generic definition of specializations
 // of the operator template are provided, nor are there any generic
@@ -867,13 +911,13 @@ inline void Atomic::release_store_fence(volatile D* p, T v) {
 template<typename D, typename I>
 inline D Atomic::add(D volatile* dest, I add_value,
                      atomic_memory_order order) {
-  return AddImpl<D, I>::add_and_fetch(dest, add_value, order);
+  return AddImpl<D, I>::add_then_fetch(dest, add_value, order);
 }
 
 template<typename D, typename I>
-inline D Atomic::fetch_and_add(D volatile* dest, I add_value,
-                               atomic_memory_order order) {
-  return AddImpl<D, I>::fetch_and_add(dest, add_value, order);
+inline D Atomic::fetch_then_add(D volatile* dest, I add_value,
+                                atomic_memory_order order) {
+  return AddImpl<D, I>::fetch_then_add(dest, add_value, order);
 }
 
 template<typename D, typename I>
@@ -884,13 +928,13 @@ struct Atomic::AddImpl<
                     (sizeof(I) <= sizeof(D)) &&
                     (std::is_signed<I>::value == std::is_signed<D>::value)>::type>
 {
-  static D add_and_fetch(D volatile* dest, I add_value, atomic_memory_order order) {
+  static D add_then_fetch(D volatile* dest, I add_value, atomic_memory_order order) {
     D addend = add_value;
-    return PlatformAdd<sizeof(D)>().add_and_fetch(dest, addend, order);
+    return PlatformAdd<sizeof(D)>().add_then_fetch(dest, addend, order);
   }
-  static D fetch_and_add(D volatile* dest, I add_value, atomic_memory_order order) {
+  static D fetch_then_add(D volatile* dest, I add_value, atomic_memory_order order) {
     D addend = add_value;
-    return PlatformAdd<sizeof(D)>().fetch_and_add(dest, addend, order);
+    return PlatformAdd<sizeof(D)>().fetch_then_add(dest, addend, order);
   }
 };
 
@@ -928,14 +972,14 @@ struct Atomic::AddImpl<
     return (P*) result;
   }
 
-  static P* add_and_fetch(P* volatile* dest, I addend, atomic_memory_order order) {
-    return scale_result(PlatformAdd<sizeof(P*)>().add_and_fetch(unscale_dest(dest),
+  static P* add_then_fetch(P* volatile* dest, I addend, atomic_memory_order order) {
+    return scale_result(PlatformAdd<sizeof(P*)>().add_then_fetch(unscale_dest(dest),
                                                                 scale_addend(addend),
                                                                 order));
   }
 
-  static P* fetch_and_add(P* volatile* dest, I addend, atomic_memory_order order) {
-    return scale_result(PlatformAdd<sizeof(P*)>().fetch_and_add(unscale_dest(dest),
+  static P* fetch_then_add(P* volatile* dest, I addend, atomic_memory_order order) {
+    return scale_result(PlatformAdd<sizeof(P*)>().fetch_then_add(unscale_dest(dest),
                                                                 scale_addend(addend),
                                                                 order));
   }
@@ -962,7 +1006,7 @@ inline bool Atomic::replace_if_null(D* volatile* dest, T* value,
   // Presently using a trivial implementation in terms of cmpxchg.
   // Consider adding platform support, to permit the use of compiler
   // intrinsics like gcc's __sync_bool_compare_and_swap.
-  D* expected_null = NULL;
+  D* expected_null = nullptr;
   return expected_null == cmpxchg(dest, expected_null, value, order);
 }
 
@@ -1171,4 +1215,18 @@ inline D Atomic::xchg(volatile D* dest, T exchange_value, atomic_memory_order or
   return XchgImpl<D, T>()(dest, exchange_value, order);
 }
 
+template<size_t byte_size>
+template<typename T>
+inline T Atomic::XchgUsingCmpxchg<byte_size>::operator()(T volatile* dest,
+                                             T exchange_value,
+                                             atomic_memory_order order) const {
+  STATIC_ASSERT(byte_size == sizeof(T));
+
+  T old_value;
+  do {
+    old_value = Atomic::load(dest);
+  } while (old_value != Atomic::cmpxchg(dest, old_value, exchange_value, order));
+  return old_value;
+}
+
 #endif // SHARE_RUNTIME_ATOMIC_HPP
